"""
metaflow._vendor is for vendoring dependencies of metaflow. Files 
inside of metaflow._vendor should be considered immutable and 
should only be updated to versions from upstream. 

This folder is generated by `python vendor.py`

If you would like to debundle the vendored dependencies, please 
reach out to the maintainers at chat.metaflow.org
"""
